# noqa: D100
from dataclasses import dataclass
from typing import Any


@dataclass
class File:
    """Represents the PDF file used as input."""

    payload: bytes
    file_name: str | None = None
    mime_type: str | None = None

    def to_tuple(self) -> tuple[str | None, bytes, str | None]:
        """Return a tuple for httpx mutlipart/form-data encoding."""
        return self.file_name, self.payload, self.mime_type


@dataclass
class Form:
    """Represents form data accepted by GROBID's processFulltextDocument endpoint."""

    file: File
    segment_sentences: bool | None = None
    consolidate_header: int | None = None
    consolidate_citations: int | None = None
    include_raw_citations: bool | None = None
    include_raw_affiliations: bool | None = None
    tei_coordinates: str | None = None

    def to_dict(self) -> dict[str, Any]:
        """Return dictionary for multipart/form-data."""
        form_dict: dict[str, Any] = {}

        form_dict["input"] = self.file.to_tuple()

        if self.segment_sentences:
            form_dict["segmentSentences"] = "1"

        match self.consolidate_header:
            case (0 | 1 | 2):
                form_dict["consolidateHeader"] = str(self.consolidate_header)

        match self.consolidate_citations:
            case (0 | 1 | 2):
                form_dict["consolidateCitations"] = str(self.consolidate_citations)

        if self.include_raw_citations is not None:
            form_dict["includeRawCitations"] = self.include_raw_citations

        if self.include_raw_affiliations is not None:
            form_dict["includeRawAffiliations"] = self.include_raw_affiliations

        if self.tei_coordinates is not None:
            form_dict["teiCoordinates"] = self.tei_coordinates

        return form_dict



"""GROBID client module.

Example::

    from pathlib import Path
    from grobid.models.form import Form, File
    from grobid.models.response import Response

    pdf_file = Path("<your-academic-article>.pdf")
    with open(pdf_file, "rb") as file:
        form = Form(
            file=File(
                payload=file.read(),
                file_name=pdf_file.name,
                mime_type="application/pdf",
            )
        )
        c = Client(base_url="<base-url>", form=form)
        try:
            xml_content = c.sync_request().content  # TEI XML file in bytes
        except GrobidClientError as e:
            print(e)

"""
from dataclasses import dataclass
from typing import Any

import httpx

from grobid.models.form import Form
from grobid.models.response import Response


class GrobidClientError(BaseException):
    """Exception for Client class."""

    pass


@dataclass
class Client:
    """Client for GROBID's processFulltextDocument endpoint."""

    base_url: str
    form: Form
    timeout: int = 15

    def __build_request(self) -> dict[str, Any]:
        """Build request dictionary."""
        # FIXME: api url is hardcoded
        url = f"{self.base_url}/api/processFulltextDocument"
        return dict(url=url, files=self.form.to_dict(), timeout=self.timeout)

    def __build_response(self, response: httpx.Response) -> Response:
        """Build Response object.

        Raises:
            httpx.HTTPError: if response has 203, 400, 503 or 500 status code
        """
        res = Response(
            status_code=response.status_code,
            content=response.content,
            headers=response.headers,
        )
        res.raise_for_status()

        return res

    async def asyncio_request(self) -> Response:
        """Request client asynchronously.

        Raises:
            GrobidClientError: if httpx.RequestError or httpx.HTTPError is raised
        """
        kwargs = self.__build_request()
        async with httpx.AsyncClient() as client:
            try:
                response = await client.post(**kwargs)
                return self.__build_response(response)
            except httpx.RequestError as exc:
                raise GrobidClientError(
                    f"An error occurred while requesting {exc.request.url!r}."
                )
            except httpx.HTTPError as exc:
                raise GrobidClientError(exc)

    def sync_request(self) -> Response:
        """Request client synchronously.

        Raises:
            GrobidClientError: if httpx.RequestError or httpx.HTTPError is raised
        """
        kwargs = self.__build_request()
        try:
            response = httpx.post(**kwargs)
            return self.__build_response(response)
        except httpx.RequestError as exc:
            raise GrobidClientError(
                f"An error occurred while requesting {exc.request.url!r}."
            )
        except httpx.HTTPError as exc:
            raise GrobidClientError(exc)



# noqa: D100
import httpx
import pytest
import respx

from grobid.client import Client, GrobidClientError
from grobid.models.form import File, Form


class TestClient:
    """Unit tests for Client class."""

    import fitz

    with fitz.open(filetype="pdf") as test_pdf:
        # Required to save pdf
        test_pdf.new_page()
        test_obj: bytes = test_pdf.tobytes()

    form = Form(
        file=File(
            payload=test_obj,
            file_name="Test",
            mime_type="application/pdf",
        )
    )

    timeout = 15

    @respx.mock
    def test_sync_request_status_codes(self):
        """Test GROBID's documented HTTP errors synchronously."""
        base_url = "http://validurl:8070"
        c = Client(base_url=base_url, form=self.form, timeout=self.timeout)

        # 203
        respx.mock.post(base_url).mock(return_value=httpx.Response(203))
        with pytest.raises(GrobidClientError, match="Content couldn't be extracted"):
            c.sync_request()

        # 400
        respx.mock.post(base_url).mock(return_value=httpx.Response(400))
        with pytest.raises(
            GrobidClientError,
            match="Wrong request, missing parameters, missing header",
        ):
            c.sync_request()

        # 500
        respx.mock.post(base_url).mock(return_value=httpx.Response(500))
        with pytest.raises(
            GrobidClientError,
            match="Internal service error",
        ):
            c.sync_request()

        # 503
        respx.mock.post(base_url).mock(return_value=httpx.Response(503))
        with pytest.raises(
            GrobidClientError,
            match="Service not available",
        ):
            c.sync_request()

        # 200
        respx.mock.post(base_url).mock(return_value=httpx.Response(200))
        assert c.sync_request().status_code == 200

    def test_sync_invalid_request(self):
        """Test invalid URL synchronously."""
        base_url = "http://invalidurl:8070"
        c = Client(base_url=base_url, form=self.form, timeout=self.timeout)
        with pytest.raises(
            GrobidClientError, match=r"An error occurred while requesting .*"
        ):
            c.sync_request()

    @respx.mock
    @pytest.mark.asyncio
    async def test_asyncio_request_status_codes(self):
        """Test GROBID's documented HTTP errors asynchronously."""
        base_url = "http://validurl:8070"
        c = Client(base_url=base_url, form=self.form, timeout=self.timeout)

        # 203
        respx.mock.post(base_url).mock(return_value=httpx.Response(203))
        with pytest.raises(GrobidClientError, match="Content couldn't be extracted"):
            await c.asyncio_request()

        # 400
        respx.mock.post(base_url).mock(return_value=httpx.Response(400))
        with pytest.raises(
            GrobidClientError,
            match="Wrong request, missing parameters, missing header",
        ):
            await c.asyncio_request()

        # 500
        respx.mock.post(base_url).mock(return_value=httpx.Response(500))
        with pytest.raises(
            GrobidClientError,
            match="Internal service error",
        ):
            await c.asyncio_request()

        # 503
        respx.mock.post(base_url).mock(return_value=httpx.Response(503))
        with pytest.raises(
            GrobidClientError,
            match="Service not available",
        ):
            await c.asyncio_request()

        # 200
        respx.mock.post(base_url).mock(return_value=httpx.Response(200))
        r = await c.asyncio_request()
        assert r.status_code == 200

    @pytest.mark.asyncio
    async def test_asyncio_invalid_request(self):
        """Test invalid URL asynchronously."""
        base_url = "http://invalidurl:8070"
        c = Client(base_url=base_url, form=self.form, timeout=self.timeout)
        with pytest.raises(
            GrobidClientError, match=r"An error occurred while requesting .*"
        ):
            await c.asyncio_request()



"""Export TEI models."""
from .article import Article, Table
from .citation import (
    Affiliation,
    Author,
    Citation,
    CitationIDs,
    Date,
    PageRange,
    PersonName,
    Scope,
)
from .section import Marker, Ref, RefText, Section

__all__ = [
    "Article",
    "PageRange",
    "Scope",
    "Date",
    "PersonName",
    "Affiliation",
    "Author",
    "CitationIDs",
    "Citation",
    "Ref",
    "RefText",
    "Marker",
    "Section",
    "Table",
]



"""GROBID annotation guidelines compliant parser module.

Example::

    from grobid.tei import Parser

    xml_content: bytes
    parser = Parser(xml_content)
    article = parser.parse()
    article.to_json()  # throws RuntimeError if extra require 'json' not installed

"""
import string
from typing import Generator

from bs4 import BeautifulSoup
from bs4.element import NavigableString, PageElement, Tag

from grobid.models import (
    Affiliation,
    Article,
    Author,
    Citation,
    CitationIDs,
    Date,
    PageRange,
    PersonName,
    Ref,
    RefText,
    Scope,
    Section,
    Table,
)
from grobid.models.section import Marker


class GrobidParserError(BaseException):
    """Exception for Parser class."""

    pass


# NOTE: shouldn't TEI methods be static?
class Parser:
    """Methods used to parse TEI XML into serializable objects."""

    soup: BeautifulSoup

    def __init__(self, stream: bytes) -> None:
        """TEI class constructor.

        Args:
            stream: XML bytes
        """
        self.soup = BeautifulSoup(stream, "lxml-xml")

    def __parse_date(self, date: str) -> Date | None:
        # Naive ISO 8601 date parser
        tokens = date.split(sep="-")
        tokens = list(filter(None, tokens))

        match len(tokens):
            case 0:
                return
            case 1:
                year = tokens[0]
                return Date(year)
            case 2:
                year, month = tokens
                return Date(year, month)
            case _:
                year, month, day = tokens[0:3]
                return Date(year, month, day)

    def __text_and_refs(
        self,
        source_tag: Tag,
    ) -> Generator[PageElement, str, None]:
        # Generator with both NavigableStrings and ref Tags
        for descendant in source_tag.descendants:
            descendant_type = type(descendant)
            if descendant_type is Tag and descendant.name == "ref":  # type: ignore
                yield descendant
            elif descendant_type is NavigableString:
                yield descendant

    def parse(self) -> Article:
        """Attempt to parse the XML into Article object.

        Parsing is strict (fails if fields are missing)

        Returns:
            Article object

        Raises:
            GrobidParserError: Article could not be parsed
        """
        body = self.soup.body

        if not isinstance(body, Tag):
            raise GrobidParserError("Missing body")

        abstract: Section | None = self.section(self.soup.abstract, title="Abstract")

        sections: list[Section] = []
        for div in body.find_all("div"):
            if (section := self.section(div)) is not None:
                sections.append(section)

        tables: dict[str, Table] = {}
        for table_tag in body.find_all("figure", {"type": "table"}):
            if isinstance(table_tag, Tag):
                if "xml:id" in table_tag.attrs:
                    name = table_tag.attrs["xml:id"]
                    if (table_obj := self.table(table_tag)) is not None:
                        tables[name] = table_obj

        if (source := self.soup.find("sourceDesc")) is None:
            raise GrobidParserError("Missing source description")

        biblstruct_tag = source.find("biblStruct")
        if not isinstance(biblstruct_tag, Tag):
            raise GrobidParserError("Missing bibliography")

        bibliography = self.citation(biblstruct_tag)
        keywords = self.keywords(self.soup.keywords)

        listbibl_tag = self.soup.find("listBibl")
        if not isinstance(listbibl_tag, Tag):
            raise GrobidParserError("Missing citations")

        citations = {}
        for struct_tag in listbibl_tag.find_all("biblStruct"):
            if isinstance(struct_tag, Tag):
                name = struct_tag.get("xml:id")
                citations[name] = self.citation(struct_tag)

        return Article(
            abstract=abstract,
            sections=sections,
            tables=tables,
            bibliography=bibliography,
            keywords=keywords,
            citations=citations,
        )

    def citation(self, source_tag: Tag) -> Citation:
        """Parse citation.

        Args:
            source_tag : biblStruct XML Tag

        Returns:
            Citation object
        """
        title = self.title(source_tag, attrs={"type": "main"})
        if not title:
            # Use meeting as the main title
            title = self.title(source_tag, attrs={"level": "m"})
        citation = Citation(title=title)
        citation.authors = self.authors(source_tag)
        ids = CitationIDs(
            DOI=self.idno(source_tag, attrs={"type": "DOI"}),
            arXiv=self.idno(source_tag, attrs={"type": "arXiv"}),
        )
        if not ids.is_empty():
            citation.ids = ids

        citation.date = self.date(source_tag)
        citation.target = self.target(source_tag)
        citation.publisher = self.publisher(source_tag)
        citation.scope = self.scope(source_tag)
        if journal := self.title(source_tag, attrs={"level": "j"}):
            if journal != citation.title:
                citation.journal = journal
        if series := self.title(source_tag, attrs={"level": "s"}):
            if series != citation.title:
                citation.series = series

        return citation

    def title(self, source_tag: Tag | None, attrs: dict[str, str] = {}) -> str:
        """Parse title tag text.

        Args:
            source_tag : XML tag
            attrs: dictionary of filters on attribute values. Default is empty dict.

        Returns:
            Text in title tag if it exists
        """
        title: str = ""
        if source_tag is not None:
            if (title_tag := source_tag.find("title", attrs=attrs)) is not None:
                title = title_tag.text

        return title

    def target(self, source_tag: Tag | None) -> str | None:
        """Parse ptr tag target.

        Args:
            source_tag : XML tag

        Returns:
            Target location in ptr tag if it exists
        """
        if source_tag is not None:
            if (ptr_tag := source_tag.ptr) is not None:
                if "target" in ptr_tag.attrs:
                    # TODO: validate URL
                    return ptr_tag.attrs["target"]

    def idno(self, source_tag: Tag | None, attrs: dict[str, str] = {}) -> str | None:
        """Parse idno tag.

        Args:
            source_tag : XML tag
            attrs: dictionary of filters on attribute values. Default is empty dict.

        Returns:
            Text content of idno_tag if it exists
        """
        if source_tag is not None:
            if (idno_tag := source_tag.find("idno", attrs=attrs)) is not None:
                return idno_tag.text or None

    def keywords(self, source_tag: Tag | None) -> set[str]:
        """Parse all term tags.

        Args:
            source_tag : XML tag

        Returns:
            Set of keywords
        """
        keywords: set[str] = set()

        if source_tag is not None:
            for term_tag in source_tag.find_all("term"):
                if term_tag.text:
                    if clean_keyword := self.clean_title_string(term_tag.text):
                        keywords.add(clean_keyword)

        return keywords

    def publisher(self, source_tag: Tag | None) -> str | None:
        """Parse publisher tag text.

        Args:
            source_tag : XML tag

        Returns:
            Text in publisher tag if it exists
        """
        if source_tag is not None:
            if (publisher_tag := source_tag.find("publisher")) is not None:
                return publisher_tag.text or None

    def date(self, source_tag: Tag | None) -> Date | None:
        """Parse date tag.

        Args:
            source_tag : XML tag

        Returns:
            Date object if date tag is valid
        """
        if source_tag is not None:
            if (date_tag := source_tag.date) is not None:
                if "when" in date_tag.attrs:
                    when = date_tag.attrs["when"]

                    return self.__parse_date(when)

    def scope(self, source_tag: Tag | None) -> Scope | None:
        """Parse all biblScope tags.

        Args:
            source_tag : XML tag

        Returns:
            Scope object if biblScope tags exist
        """
        if source_tag is not None:
            scope = Scope()
            for scope_tag in source_tag.find_all("biblScope"):
                match scope_tag.attrs["unit"]:
                    case "page":
                        try:
                            if "from" in scope_tag.attrs and "to" in scope_tag.attrs:
                                from_page = int(scope_tag["from"])
                                to_page = int(scope_tag["to"])
                            elif scope_tag.text:
                                from_page = int(scope_tag.text)
                                to_page = from_page
                            else:
                                continue

                            scope.pages = PageRange(
                                from_page=from_page, to_page=to_page
                            )
                        except ValueError:
                            continue
                    case "volume":
                        try:
                            volume = int(scope_tag.text)
                            scope.volume = volume
                        except ValueError:
                            continue

            if not scope.is_empty():
                return scope

    def authors(self, source_tag: Tag | None) -> list[Author]:
        """Parse all author tags.

        Args:
            source_tag : XML tag

        Returns:
            List of Author objects
        """
        authors: list[Author] = []
        if source_tag is not None:
            for author in source_tag.find_all("author"):
                author_obj: Author | None = None
                if (persname := author.find("persName")) is not None:
                    if (surname_tag := persname.find("surname")) is not None:
                        person_name = PersonName(surname=surname_tag.text)
                        if forename_tag := persname.find("forename", {"type": "first"}):
                            person_name.first_name = forename_tag.text

                        author_obj = Author(person_name=person_name)
                        authors.append(author_obj)

                        if email_tag := author.find("email"):
                            author_obj.email = email_tag.text

                        for affiliation_tag in author.find_all("affiliation"):
                            affiliation_obj = Affiliation()
                            for orgname_tag in affiliation_tag.find_all("orgName"):
                                match orgname_tag["type"]:
                                    case "institution":
                                        affiliation_obj.institution = orgname_tag.text
                                    case "department":
                                        affiliation_obj.department = orgname_tag.text
                                    case "laboratory":
                                        affiliation_obj.laboratory = orgname_tag.text

                            if not affiliation_obj.is_empty():
                                author_obj.affiliations.append(affiliation_obj)

        return authors

    def section(self, source_tag: Tag | None, title: str = "") -> Section | None:
        """Parse div tag with head tag.

        Capitalizes title if not already.

        Section can have an empty body.

        Args:
            source_tag : XML tag
            title: forces the parsing of the section. Default is empty string (false)

        Returns:
            Section object if valid section.
        """
        if source_tag is not None:
            head = source_tag.find("head")
            if isinstance(head, Tag):
                head_text: str = head.get_text()
                if "n" in head.attrs or head_text[0] in string.ascii_letters:
                    if head_text.isupper() or head_text.islower():
                        head_text = head_text.capitalize()

                section = Section(title=head_text)
            elif title:
                section = Section(title=title)
            else:
                return

            paragraphs = source_tag.find_all("p")
            for p in paragraphs:
                if p and (ref_text := self.ref_text(p)) is not None:
                    section.paragraphs.append(ref_text)

            return section

    def ref_text(self, source_tag: Tag | None) -> RefText | None:
        """Parse text with ref tags.

        Args:
            source_tag : XML tag

        Returns:
            RefText object
        """
        if source_tag is not None:
            text_and_refs = self.__text_and_refs(source_tag)
            start = 0
            ref_text = RefText(text="")
            for el in text_and_refs:
                start = len(ref_text.text)
                if isinstance(el, Tag):
                    end = start + len(el.text)
                    ref = Ref(start=start, end=end)
                    if (el_type := el.attrs.get("type")) is not None:
                        try:
                            ref.marker = Marker[el_type]
                        except KeyError:
                            pass

                    # NOTE: if target[0] is '#', check for citation
                    if (el_target := el.attrs.get("target")) is not None:
                        ref.target = el_target

                    ref_text.refs.append(ref)
                else:
                    ref_text.text += str(el)

            return ref_text

    def table(self, source_tag: Tag | None) -> Table | None:
        """Parse <figure> with table type.

        Args:
            source_tag : XML tag

        Returns:
            Table object
        """
        if source_tag is not None:
            if (head_tag := source_tag.find("head")) is not None:
                if head_text := head_tag.get_text():
                    table = Table(heading=head_text)
                    if (desc_tag := source_tag.find("figDesc")) is not None:
                        table.description = desc_tag.get_text()
                    rows = source_tag.find_all("row")
                    for row in rows:
                        row_list = []
                        for cell in row.find_all("cell"):
                            row_list.append(cell.get_text())
                        table.rows.append(row_list)

                    return table

    @staticmethod
    def clean_title_string(s: str) -> str:
        """Remove non-alpha leading characters from string.

        Args:
            s : title string

        Returns:
            Clean title string
        """
        s = s.strip()

        while s and not s[0].isalpha():
            s = s[1:]

        return s.capitalize()



"""Represents the text sections in a scholarly article."""
from dataclasses import dataclass, field

try:
    from mashumaro.mixins.json import DataClassJSONMixin
except ImportError:
    from grobid.models.misc import DataClassJSONMixin

from enum import Enum


class Marker(str, Enum):
    """Represents the callouts to structures.

    <https://grobid.readthedocs.io/en/latest/training/fulltext/#markers-callouts-to-structures>
    """

    bibr = "bibr"
    figure = "figure"
    table = "table"
    box = "box"
    formula = "formula"


@dataclass
class Ref(DataClassJSONMixin):
    """Represents <ref> XML tag.

    Stores the start and end positions of the reference rather than the text.
    """

    start: int
    end: int
    marker: Marker | None = None
    target: str | None = None


@dataclass
class RefText(DataClassJSONMixin):
    """Represents the <p> XML tag.

    Supports embedded <ref> XML tags.
    """

    text: str
    refs: list[Ref] = field(default_factory=list)

    @property
    def plain_text(self) -> str:
        """Return text without any references.

        Trailing whitespace is removed.
        """
        if len(self.refs) == 0:
            return self.text

        ranges = [(ref.start, ref.end) for ref in self.refs]
        text = ""
        left_bound = 0
        for start, end in ranges:
            text += self.text[left_bound:start].rstrip()
            left_bound = end
        text += self.text[ranges[-1][1] :].rstrip()
        return text


@dataclass
class Section(DataClassJSONMixin):
    """Represents <div> tag with <head> tag."""

    title: str
    paragraphs: list[RefText] = field(default_factory=list)

    def to_str(self) -> str:
        """Return paragraphs in plain text format."""
        text = ""
        for paragraph in self.paragraphs:
            text += paragraph.plain_text

        return text



"""Unit tests for grobid form module."""


import fitz

from grobid.models.form import File, Form

with fitz.open(filetype="pdf") as test_pdf:
    test_pdf.new_page()
    test_obj = test_pdf.tobytes()


class TestForm:
    """Unit tests for Form class."""

    file: File = File(test_obj)

    def test_no_opt_params(self):  # noqa: D102
        form: Form = Form(self.file)

        assert form.to_dict() == dict(input=self.file.to_tuple())

    def test_segment_sentences_param(self):  # noqa: D102
        form: Form = Form(self.file, segment_sentences=True)

        assert form.to_dict() == dict(input=self.file.to_tuple(), segmentSentences="1")

    def test_consolidate_header_param(self):  # noqa: D102
        form: Form = Form(self.file, consolidate_header=0)

        assert form.to_dict() == dict(input=self.file.to_tuple(), consolidateHeader="0")

    def test_consolidate_citations_param(self):  # noqa: D102
        form: Form = Form(self.file, consolidate_citations=0)

        assert form.to_dict() == dict(
            input=self.file.to_tuple(), consolidateCitations="0"
        )

    def test_include_raw_citations_param(self):  # noqa: D102
        form: Form = Form(self.file, include_raw_citations=True)

        assert form.to_dict() == dict(
            input=self.file.to_tuple(), includeRawCitations=True
        )

    def test_include_raw_affliations_param(self):  # noqa: D102
        form: Form = Form(self.file, include_raw_affiliations=True)

        assert form.to_dict() == dict(
            input=self.file.to_tuple(), includeRawAffiliations=True
        )

    def test_tei_coordinates_param(self):  # noqa: D102
        tei_coordinates = "s"
        form: Form = Form(self.file, tei_coordinates=tei_coordinates)

        assert form.to_dict() == dict(
            input=self.file.to_tuple(), teiCoordinates=tei_coordinates
        )


class TestFile:
    """Unit tests for File class."""

    def test_no_opt_params(self):  # noqa: D102
        file: File = File(test_obj)

        assert file.to_tuple() == (None, test_obj, None)

    def test_file_name_param(self):  # noqa: D102
        file_name = "test"
        file: File = File(test_obj, file_name=file_name)

        assert file.to_tuple() == (file_name, test_obj, None)

    def test_mime_type_param(self):  # noqa: D102
        mime_type = "application/pdf"
        file: File = File(test_obj, mime_type=mime_type)

        assert file.to_tuple() == (None, test_obj, mime_type)

    def test_all_params(self):  # noqa: D102
        file_name = "test"
        mime_type = "application/pdf"
        file: File = File(test_obj, file_name=file_name, mime_type=mime_type)

        assert file.to_tuple() == (file_name, test_obj, mime_type)



# noqa: D100
class DataClassDictMixin:
    """Mock for mashumaro.DataClassDictMixin."""

    def to_json(self):
        """Raises RuntimeWarning if user tries to use this method."""
        raise RuntimeWarning(
            "mashumaro is not installed. Run 'pip install grobid_client[json]'"
        )

    @classmethod
    def from_json(cls):
        """Raises RuntimeWarning if user tries to use this method."""
        raise RuntimeWarning(
            "mashumaro is not installed. Run 'pip install grobid_client[json]'"
        )


class DataClassJSONMixin(DataClassDictMixin):
    """Mock for mashumaro.DataClassJSONMixin."""

    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['grobid', 'grobid.models']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2', 'httpx>=0.23.0,<0.24.0', 'lxml>=4.9.1,<5.0.0']

extras_require = \
{'json': ['mashumaro>=3.0.3,<4.0.0']}

setup_kwargs = {
    'name': 'grobid',
    'version': '0.1.3',
    'description': 'Python library for serializing GROBID TEI XML to dataclass',
    'long_description': '# grobid\n> Python library for serializing GROBID TEI XML to [dataclasses](https://docs.python.org/3/library/dataclasses.html)\n\n[![Build Status](https://github.com/ram02z/grobid/workflows/tests/badge.svg)](https://github.com/ram02z/grobid/actions)\n[![Coverage Status](https://coveralls.io/repos/github/ram02z/grobid/badge.svg)](https://coveralls.io/github/ram02z/grobid)\n[![Latest Version](https://img.shields.io/pypi/v/grobid.svg)](https://pypi.python.org/pypi/grobid)\n[![Python Version](https://img.shields.io/pypi/pyversions/grobid.svg)](https://pypi.python.org/pypi/grobid)\n[![License](https://img.shields.io/badge/MIT-blue.svg)](https://opensource.org/licenses/MIT)\n\n## Installation\n\nUse `pip` to install:\n\n```shell\n$ pip install grobid\n$ pip install grobid[json] # for JSON serializable dataclass objects\n```\n\n\nYou can also download the `.whl` file from the release section:\n\n```shell\n$ pip install *.whl\n```\n\n## Usage\n\n### Client\n\nIn order to convert an academic PDF to TEI XML file, we use GROBID\'s REST\nservices. Specifically the [processFulltextDocument](https://grobid.readthedocs.io/en/latest/Grobid-service/#apiprocessfulltextdocument) endpoint.\n\n\n```python\nfrom pathlib import Path\nfrom grobid.models.form import Form, File\nfrom grobid.models.response import Response\n\npdf_file = Path("<your-academic-article>.pdf")\nwith open(pdf_file, "rb") as file:\n    form = Form(\n        file=File(\n            payload=file.read(),\n            file_name=pdf_file.name,\n            mime_type="application/pdf",\n        )\n    )\n    c = Client(base_url="<base-url>", form=form)\n    try:\n        xml_content = c.sync_request().content  # TEI XML file in bytes\n    except GrobidClientError as e:\n        print(e)\n```\n\nwhere `base-url` is the URL of the GROBID REST service\n\n> You can use `https://cloud.science-miner.com/grobid/` to test\n\n#### [Form](https://github.com/ram02z/grobid/blob/master/src/grobid/models/form.py#L20)\n\nThe `Form` class supports most of the optional parameters of the processFulltextDocument\nendpoint.\n\n\n### Parser\n\nIf you want to serialize the XML content, we can use the `Parser` class to\ncreate [dataclasses](https://docs.python.org/3/library/dataclasses.html)\nobjects.\n\nNot all of the GROBID annoation guidelines are met, but compliance is a goal.\nSee [#1](https://github.com/ram02z/grobid/issues/1).\n\n```python\nfrom grobid.tei import Parser\n\nxml_content: bytes\nparser = Parser(xml_content)\narticle = parser.parse()\narticle.to_json()  # throws RuntimeError if extra require \'json\' not installed\n```\n\nwhere `xml_content` is the same as in [Client section](#client)\n\nAlternately, you can load the XML from a file:\n\n```python\nfrom grobid.tei import Parser\n\nwith open("<your-academic-article>.xml", "rb") as xml_file:\n  xml_content = xml_file.read()\n  parser = Parser(xml_content)\n  article = parser.parse()\n  article.to_json()  # throws RuntimeError if extra require \'json\' not installed\n```\n\nWe use [mashumaro](https://github.com/Fatal1ty/mashumaro) to serialize the\ndataclasses into JSON (mashumaro supports other formats, you can submit a PR if\nyou want). By default, mashumaro isn\'t installed, use `pip install\ngrobid[json]`.\n\n## License\n\nMIT\n\n## Contributing\n\nYou are welcome to add missing features by submitting a PR, however, I won\'t be\naccepting any requests other than GROBID annotation compliance.\n\n## Disclaimer\n\nThis module was originally part of a group university project, however, all the\ncode and tests was also authored by me.\n',
    'author': 'Omar Zeghouani',
    'author_email': 'omarzeghouanii@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ram02z/grobid',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Unit tests for the citation module."""
from grobid.models import PersonName


class TestPersonName:
    """Unit tests for PersonName class."""

    first_name = "Foo"
    surname = "Bar"

    def test_to_string(self):
        """Tests surname only then adds first name."""
        person_name = PersonName(surname=self.surname)
        assert person_name.to_string() == self.surname

        person_name.first_name = self.first_name
        assert person_name.to_string() == f"{self.first_name} {self.surname}"



# noqa: D100
from dataclasses import dataclass

from httpx import Headers, HTTPError


@dataclass
class Response:
    """Represents the response from GROBID's processFulltextDocument endpoint."""

    status_code: int
    content: bytes
    headers: Headers

    def raise_for_status(self):
        """Only considers GROBID's documented status codes as HTTP errors."""
        http_error_msg = ""

        match self.status_code:
            case 203:
                error_msg = "Content couldn't be extracted"
            case 400:
                error_msg = "Wrong request, missing parameters, missing header"
            case 500:
                error_msg = "Internal service error"
            case 503:
                error_msg = "Service not available"
            case _:
                return

        http_error_msg = f"{self.status_code}: {error_msg}"
        raise HTTPError(http_error_msg)



"""Unit tests for the TEI class."""
import pytest

from grobid.models import (
    Affiliation,
    Article,
    Author,
    Citation,
    CitationIDs,
    Date,
    Marker,
    PageRange,
    PersonName,
    Ref,
    RefText,
    Scope,
    Section,
    Table,
)
from grobid.tei import GrobidParserError, Parser


class TestParse:
    """Unit tests for parse function."""

    @staticmethod
    def build_xml(article: Article) -> bytes:
        """Create XML from Article object."""
        tei_tags: list[bytes] = [b"<TEI>"]

        tei_tags.append(b"<sourceDesc>")
        tei_tags.append(TestCitation.build_xml(article.bibliography))
        tei_tags.append(b"</sourceDesc>")

        tei_tags.append(b"<body>")
        for section in article.sections:
            tei_tags.append(TestSection.build_xml(section))
        for xml_id, table in article.tables.items():
            tei_tags.append(TestTable.build_xml(table, xml_id))
        tei_tags.append(b"</body>")

        tei_tags.append(b"<listBibl>")
        for xml_id, citation in article.citations.items():
            tei_tags.append(TestCitation.build_xml(citation, xml_id=xml_id))
        tei_tags.append(b"</listBibl>")

        tei_tags.append(b"</TEI>")

        return b"".join(tei_tags)

    def test_no_body(self):  # noqa: D102
        xml = b"<TEI></TEI>"

        with pytest.raises(GrobidParserError, match="Missing body"):
            Parser(xml).parse()

    def test_no_sourcedesc(self):  # noqa: D102
        xml = b"<TEI><body></body></TEI>"

        with pytest.raises(GrobidParserError, match="Missing source description"):
            Parser(xml).parse()

    def test_no_biblstruct(self):  # noqa: D102
        xml = b"<TEI><sourceDesc></sourceDesc><body></body></TEI>"

        with pytest.raises(GrobidParserError, match="Missing bibliography"):
            Parser(xml).parse()

    def test_no_listbibl(self):  # noqa: D102
        xml = b"""
        <TEI><sourceDesc><biblStruct></biblStruct></sourceDesc><body></body></TEI>
        """

        with pytest.raises(GrobidParserError, match="Missing citations"):
            Parser(xml).parse()

    def test_valid_article(self):  # noqa: D102
        article = Article(
            bibliography=Citation(
                title="Test",
                authors=[
                    Author(
                        PersonName("Doe", "John"),
                    )
                ],
            ),
            keywords=set(),
            tables=dict(
                test=Table(
                    heading="Test", description="Lorem Ipsum", rows=[["Foo", "Bar"]]
                )
            ),
            sections=[Section("Introduction", [RefText("Lorem Ipsum")])],
            citations=dict(
                test=Citation(
                    title="Test2",
                    authors=[Author(PersonName("Doe", "Jane"))],
                )
            ),
        )

        tei = Parser(TestParse.build_xml(article))

        assert tei.parse() == article


class TestTitle:
    """Unit tests for the title function."""

    def test_valid_tag(self):  # noqa: D102
        title = "Test"
        xml = bytes(f"<title>{title}</title>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.title(tei.soup) == title

    def test_valid_title_attr(self):  # noqa: D102
        title = "Test"
        xml = bytes(
            f"<div><title><Invalid</title><title type='main'>{title}</></div>",
            encoding="utf-8",
        )
        tei = Parser(xml)

        assert tei.title(tei.soup, attrs={"type": "main"}) == title

    def test_invalid_tag(self):  # noqa: D102
        xml = b"<nottitle>Invalid</nottitle>"
        tei = Parser(xml)

        # NOTE: should it return empty string over None?
        assert tei.title(tei.soup) == ""


class TestTarget:
    """Unit tests for the target function."""

    def test_valid_tag(self):  # noqa: D102
        target = "http://avalidtarget.org"
        xml = bytes(f"<ptr target='{target}'/>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.target(tei.soup) == target

    def test_no_target(self):  # noqa: D102
        xml = b"<ptr/>"
        tei = Parser(xml)

        assert tei.target(tei.soup) is None

    def test_empty_tag(self):  # noqa: D102
        xml = b"<ptr/>"
        tei = Parser(xml)

        assert tei.target(tei.soup) is None


class TestIdno:
    """Unit tests for the idno function."""

    def test_valid_tag(self):  # noqa: D102
        idno = "test"
        xml = bytes(f"<idno>{idno}</idno>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.idno(tei.soup) == idno

    def test_valid_idno_attr(self):
        """Checks if attribute dictionary search is prioritised over empty tags."""
        idno = "test"
        type_ = "DOI"
        xml = bytes(
            f"<div><idno>Invalid</idno><idno type='{type_}'>{idno}</idno></div>",
            encoding="utf-8",
        )
        tei = Parser(xml)

        assert tei.idno(tei.soup, attrs={"type": type_}) == idno

    def test_empty_tag(self):  # noqa: D102
        xml = bytes("<idno></idno>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.idno(tei.soup) is None


class TestPublisher:
    """Unit tests for the publisher function."""

    def test_valid_tag(self):  # noqa: D102
        publisher = "Foo Bar"
        xml = bytes(f"<publisher>{publisher}</publisher>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.publisher(tei.soup) == publisher

    def test_empty_tag(self):  # noqa: D102
        xml = bytes("<publisher></publisher>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.publisher(tei.soup) is None


class TestKeywords:
    """Unit tests for the keywords function."""

    def test_valid_tags(self):  # noqa: D102
        keywords = {"Keywords", "Tags"}  # keywords are nouns
        term_tags: list[bytes] = [b"<keywords>"]
        for keyword in keywords:
            term_tags.append(bytes(f"<term>{keyword}</term>", encoding="utf-8"))

        term_tags.append(b"</keywords>")

        xml = b"".join(term_tags)
        tei = Parser(xml)

        assert tei.keywords(tei.soup) == keywords

    def test_empty_tag(self):  # noqa: D102
        xml = b"<term></term>"
        tei = Parser(xml)

        assert tei.keywords(tei.soup) == set()


class TestDate:
    """Unit tests for the date function."""

    def test_valid_tag_year(self):
        """Date with only the year."""
        year = "2022"
        date = Date(year=year)
        xml = bytes(f"<date when='{year}'/>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.date(tei.soup) == date

    def test_valid_tag_year_month(self):
        """Date with only the year and month."""
        year = "2022"
        month = "05"
        date = Date(year=year, month=month)
        xml = bytes(f"<date when='{year}-{month}'/>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.date(tei.soup) == date

    def test_valid_tag_full_date(self):
        """Date with year, month and day."""
        year = "2022"
        month = "05"
        day = "03"
        date = Date(year=year, month=month, day=day)
        xml = bytes(f"<date when='{year}-{month}-{day}'/>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.date(tei.soup) == date

    def test_empty_tag(self):  # noqa: D102
        xml = b"<date when=''/>"
        tei = Parser(xml)

        assert tei.date(tei.soup) is None


class TestCitation:
    """Unit tests for the citation function."""

    @staticmethod
    def build_xml(citation: Citation, xml_id: str | None = None) -> bytes:
        """Create XML from Citation object."""
        bibl_tags: list[bytes] = []
        if xml_id:
            bibl_tags.append(bytes(f"<biblStruct xml:id='{xml_id}'>", encoding="utf-8"))
        else:
            bibl_tags.append(b"<biblStruct>")

        bibl_tags.append(
            bytes(f"<title type='main'>{citation.title}</title>", encoding="utf-8")
        )

        bibl_tags.append(TestAuthors.build_xml(citation.authors))

        if citation.ids is not None:
            for k, v in citation.ids.__dict__.items():
                if v is None:
                    continue
                bibl_tags.append(
                    bytes(f"<idno type='{k}'>{v}</idno>", encoding="utf-8")
                )

        if citation.scope is not None:
            for k, v in citation.scope.__dict__.items():
                if v is None:
                    continue
                bibl_tags.append(
                    bytes(f"<biblScope unit='{k}'>{v}</biblScope>", encoding="utf-8")
                )

        if citation.target:
            bibl_tags.append(
                bytes(f"<ptr target='{citation.target}' />", encoding="utf-8")
            )

        if citation.publisher:
            bibl_tags.append(
                bytes(f"<publisher>{citation.publisher}</publisher>", encoding="utf-8")
            )

        if citation.series:
            bibl_tags.append(
                bytes(f"<title level='s'>{citation.series}</title>", encoding="utf-8")
            )

        if citation.journal:
            bibl_tags.append(
                bytes(f"<title level='j'>{citation.journal}</title>", encoding="utf-8")
            )

        bibl_tags.append(b"</biblStruct>")

        return b"".join(bibl_tags)

    def test_valid_tags(self):  # noqa: D102
        citation = Citation(
            title="Test",
            authors=[Author(PersonName(surname="Doe", first_name="John"))],
            ids=CitationIDs(DOI="10.1000/182", arXiv="arxivID"),
            target="http://citationtarget.org",
            publisher="FooBar",
            journal="Baz",
            series="Qux",
            scope=Scope(volume=1),
        )

        tei = Parser(TestCitation.build_xml(citation))

        assert tei.citation(tei.soup) == citation

        citation = Citation(
            title="Test2",
            authors=[Author(PersonName(surname="Doe", first_name="Jane"))],
            ids=CitationIDs(DOI="10.1000/183"),
            target="http://citationtarget.org",
            publisher="FooBar",
            scope=Scope(volume=1),
        )

        tei = Parser(TestCitation.build_xml(citation))

        assert tei.citation(tei.soup) == citation


class TestScope:
    """Unit tests for the scope function."""

    def test_valid_tag_volume(self):  # noqa: D102
        volume = 7
        scope = Scope(volume=volume)
        xml = bytes(f"<biblScope unit='volume'>{volume}</biblScope>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.scope(tei.soup) == scope

    def test_valid_tag_page(self):  # noqa: D102
        page = 1
        scope = Scope(pages=PageRange(page, page))
        xml = bytes(f"<biblScope unit='page'>{page}</biblScope>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.scope(tei.soup) == scope

    def test_valid_tag_page_range(self):  # noqa: D102
        from_page, to_page = 1, 2
        scope = Scope(pages=PageRange(from_page, to_page))
        xml = bytes(
            f"<biblScope unit='page' from='{from_page}' to='{to_page}'>",
            encoding="utf-8",
        )
        tei = Parser(xml)

        assert tei.scope(tei.soup) == scope

    def test_invalid_page_type(self):
        """Page should be of int."""
        page = "one"
        xml = bytes(f"<biblScope unit='page'>{page}</biblScope>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.scope(tei.soup) is None

    def test_empty_page(self):  # noqa: D102
        xml = b"<biblScope unit='page'></biblScope>"
        tei = Parser(xml)

        assert tei.scope(tei.soup) is None

    def test_invalid_volume_type(self):
        """Volume should be of int."""
        volume = "seven"
        xml = bytes(f"<biblScope unit='volume'>{volume}</biblScope>", encoding="utf-8")
        tei = Parser(xml)

        assert tei.scope(tei.soup) is None


class TestAuthors:
    """Unit tests for the authors function."""

    @staticmethod
    def build_xml(authors: list[Author]) -> bytes:
        """Create XML from list of author objects."""
        author_tags: list[bytes] = [b"<analytic>"]
        for author in authors:
            author_tags.append(b"<author>")
            author_tags.append(b"<persName>")
            if (first_name := author.person_name.first_name) is not None:
                author_tags.append(
                    bytes(
                        f"<forename type='first'>{first_name}</forename>",
                        encoding="utf-8",
                    )
                )
            author_tags.append(
                bytes(
                    f"<surname>{author.person_name.surname}</surname>",
                    encoding="utf-8",
                )
            )

            author_tags.append(b"</persName>")

            if (email := author.email) is not None:
                author_tags.append(bytes(f"<email>{email}</email>", encoding="utf-8"))

            author_tags.append(b"<affiliation>")
            for affiliation in author.affiliations:
                for k, v in affiliation.__dict__.items():
                    if v is None:
                        continue
                    author_tags.append(
                        bytes(f"<orgName type='{k}'>{v}</orgName>", encoding="utf-8")
                    )
            author_tags.append(b"</affiliation>")
            author_tags.append(b"</author>")

        author_tags.append(b"</analytic>")

        return b"".join(author_tags)

    def test_valid_tags(self):
        """Tests all supported tags."""
        authors = [
            Author(
                PersonName(surname="Smith", first_name="Jane"),
                email="jamesmith@gmail.com",
                affiliations=[Affiliation(department="English")],
            ),
            Author(
                PersonName(surname="Doe", first_name="John"),
                affiliations=[Affiliation(institution="University of Nottingham")],
            ),
            Author(
                PersonName(surname="Schmoe", first_name="Joe"),
                affiliations=[Affiliation(laboratory="Computer Lab")],
            ),
        ]

        tei = Parser(TestAuthors.build_xml(authors))

        assert tei.authors(tei.soup) == authors


class TestSection:
    """Unit tests for section function."""

    @staticmethod
    def build_xml(section: Section) -> bytes:
        """Create XML from Section object."""
        div_tags: list[bytes] = [b"<div>"]

        div_tags.append(bytes(f"<head>{section.title}</title>", encoding="utf-8"))

        for p in section.paragraphs:
            text_xml = p.text
            for ref in p.refs:
                marker = None
                if ref.marker:
                    marker = ref.marker.name
                ref_xml = f"<ref type='{marker}' target='{ref.target}'>{p.text[ref.start:ref.end]}</ref>"  # noqa: E501
                text_xml = text_xml[: ref.start] + ref_xml + text_xml[ref.end :]
            div_tags.append(bytes(f"<p>{text_xml}</p>", encoding="utf-8"))

        div_tags.append(b"</div>")

        return b"".join(div_tags)

    def test_valid_tag(self):  # noqa: D102
        text = "Lorem ipsum"
        section = Section(
            title="test",
            paragraphs=[
                RefText(
                    text=f"{text} [1]",
                    refs=[Ref(start=12, end=15, target="#1", marker=Marker.bibr)],
                )
            ],
        )

        for paragraph in section.paragraphs:
            assert paragraph.plain_text == text

        assert section.to_str() == text

        xml = TestSection.build_xml(section)
        tei = Parser(xml)
        # NOTE: woraround for forced capitalisation
        section.title = section.title.capitalize()

        assert tei.section(tei.soup) == section

    def test_valid_tag_no_ref(self):
        """Test for text that doesn't contain a reference."""
        text = "Lorem ipsum"
        section = Section(
            title="test",
            paragraphs=[
                RefText(
                    text=text,
                    refs=[],
                )
            ],
        )

        for paragraph in section.paragraphs:
            assert paragraph.plain_text == text

        assert section.to_str() == text

        xml = TestSection.build_xml(section)
        tei = Parser(xml)
        # NOTE: woraround for forced capitalisation
        section.title = section.title.capitalize()

        assert tei.section(tei.soup) == section

    def test_valid_tag_invalid_ref(self):
        """Test for text that contains a ref with an invalid marker (None)."""
        text = "Lorem ipsum"
        section = Section(
            title="test",
            paragraphs=[
                RefText(
                    text=f"{text} [1]",
                    refs=[Ref(start=12, end=15, target="#1", marker=None)],
                )
            ],
        )

        for paragraph in section.paragraphs:
            assert paragraph.plain_text == text

        assert section.to_str() == text

        xml = TestSection.build_xml(section)
        tei = Parser(xml)
        # NOTE: woraround for forced capitalisation
        section.title = section.title.capitalize()

        assert tei.section(tei.soup) == section

    def test_title_tag(self):
        """Test for divs that don't have <head> tag (i.e. <abstract>)."""
        title = "abstract"
        xml = bytes(f"<{title}></{title}>", encoding="utf-8")
        tei = Parser(xml)

        section = tei.section(tei.soup, title=title)

        assert section is not None
        assert section.title == title

    def test_empty_tag(self):  # noqa: D102
        xml = b"<div></div>"
        tei = Parser(xml)

        assert tei.section(tei.soup) is None


class TestTable:
    """Unit tests for table function."""

    @staticmethod
    def build_xml(table: Table, xml_id: str | None = None) -> bytes:
        """Create XML from Table object."""
        figure_tags: list[bytes] = []
        if xml_id:
            figure_tags.append(
                bytes(f"<figure xml:id='{xml_id}' type='table'>", encoding="utf-8")
            )
        else:
            figure_tags.append(b"<figure type='table'>")

        figure_tags.append(bytes(f"<head>{table.heading}</head>", encoding="utf-8"))

        if table.description is not None:
            figure_tags.append(
                bytes(f"<figDesc>{table.description}</figDesc>", encoding="utf-8")
            )

        for row in table.rows:
            figure_tags.append(b"<row>")
            for cell in row:
                figure_tags.append(bytes(f"<cell>{cell}</cell>", encoding="utf-8"))
            figure_tags.append(b"</row>")

        figure_tags.append(b"</figure>")

        return b"".join(figure_tags)

    def test_valid_tag(self):  # noqa: D102
        table = Table(heading="Test", description="Lorem Ipsum", rows=[["Foo", "Bar"]])
        xml = TestTable.build_xml(table)

        tei = Parser(xml)

        assert tei.table(tei.soup) == table

    def test_empty_tag(self):
        """Empty head tag."""
        xml = b"<figure><head></head></figure>"
        tei = Parser(xml)

        assert tei.table(tei.soup) is None


class TestCleanTitleString:
    """Unit tests for the clean_title_string function."""

    def test_non_alpha_string(self):  # noqa: D102
        s = "123"

        assert Parser.clean_title_string(s) == ""

    def test_clean_alpha_string(self):  # noqa: D102
        s = "Test"

        assert Parser.clean_title_string(s) == s

    def test_clean_dirty_string(self):  # noqa: D102
        s = "Test"
        dirty = f"21 {s}"

        assert Parser.clean_title_string(dirty) == s



# noqa: D100
from dataclasses import dataclass, field

try:
    from mashumaro.mixins.json import DataClassJSONMixin
except ImportError:
    from grobid.models.misc import DataClassJSONMixin

from grobid.models.citation import Citation
from grobid.models.section import Section


@dataclass
class Table(DataClassJSONMixin):
    """Represents the <figure> XML tag of type table."""

    heading: str
    description: str | None = None
    rows: list[list[str]] = field(default_factory=list)


@dataclass
class Article(DataClassJSONMixin):
    """Represents the scholarly article."""

    bibliography: Citation
    keywords: set[str]
    citations: dict[str, Citation]
    sections: list[Section]
    tables: dict[str, Table]
    abstract: Section | None = None



"""Represents a citation, including all the relevant information."""
from dataclasses import dataclass, field

try:
    from mashumaro.mixins.json import DataClassJSONMixin
except ImportError:
    from grobid.models.misc import DataClassJSONMixin


@dataclass
class PageRange(DataClassJSONMixin):
    """Represents the 'to' and 'from' attributes in <biblScope/> XML tag."""

    from_page: int
    to_page: int


@dataclass
class Scope(DataClassJSONMixin):
    """Represents the <biblScope/> XML tag."""

    volume: int | None = None
    pages: PageRange | None = None

    def is_empty(self) -> bool:
        """Return True if the default values are the same."""
        return all(attr is None for attr in self.__dict__.values())


@dataclass
class Date(DataClassJSONMixin):
    """Represents the 'when' attribute in the <date/> XML tag."""

    year: str
    month: str | None = None
    day: str | None = None


@dataclass
class PersonName(DataClassJSONMixin):
    """Represents the <persName/> XML tag."""

    surname: str
    first_name: str | None = None
    # middle_name: str | None = None
    # title: str | None = None

    def to_string(self) -> str:
        """Return string representation of object."""
        if self.first_name:
            return f"{self.first_name} {self.surname}"
        else:
            return f"{self.surname}"


@dataclass
class Affiliation(DataClassJSONMixin):
    """Represents the <affiliation> XML tag."""

    # key: str
    department: str | None = None
    institution: str | None = None
    laboratory: str | None = None
    # address: Address | None = None

    def is_empty(self) -> bool:
        """Return True if the default values are the same."""
        return all(attr is None for attr in self.__dict__.values())


@dataclass
class Author(DataClassJSONMixin):
    """Represents the <author> XML tag."""

    person_name: PersonName
    affiliations: list[Affiliation] = field(default_factory=list)
    email: str | None = None


@dataclass
class CitationIDs(DataClassJSONMixin):
    """Represents the <idno> XML tag."""

    DOI: str | None = None
    arXiv: str | None = None  # noqa: N815
    # issn: str | None = None
    # pii: str | None = None
    # other: str | None = None

    def is_empty(self) -> bool:
        """Return True if the default values are the same."""
        return all(attr is None for attr in self.__dict__.values())


@dataclass
class Citation(DataClassJSONMixin):
    """Represents the <biblStruct> XML tag."""

    title: str
    authors: list[Author] = field(default_factory=list)
    date: Date | None = None
    ids: CitationIDs | None = None
    target: str | None = None
    publisher: str | None = None
    journal: str | None = None
    series: str | None = None
    scope: Scope | None = None
    # meeting: str | None = None
    # phone: str | None = None
